MAX_EMBED_TITLE_LENGTH = 256
MAX_EMBED_DESCRIPTION_LENGTH = 2048
MAX_EMBED_FOOTER_TEXT_LENGTH = 2048
MAX_EMBED_AUTHOR_NAME_LENGTH = 256
MAX_EMBED_TOTAL_LENGTH = 6000
MAX_EMBED_FIELD_NAME_LENGTH = 256
MAX_EMBED_FIELD_VALUE_LENGTH = 1024
MAX_EMBED_FIELD_COUNT = 25